from itertools import count


def Result(request, result_class=dict):
    return result_class(request.client, request.get())


class ResultList(object):
    limit = 20

    def __init__(self, request, result_class=dict):
        self.request = request
        self.result_class = result_class
        self.cache = {}

    def _stop(self, offset, limit, c, step):
        return offset + limit > c or (step and offset + step > c)

    def _result(self, data):
        return self.result_class(self.request.client, data)

    def __getitem__(self, key):
        if isinstance(key, slice):
            xargs = []
            start = key.start if key.start else 0
            if start < 0:
                raise IndexError('Negative indices are not supported')
            xargs.append(start)
            if key.stop:
                iterator_class = xrange
                if key.stop < 0:
                    raise IndexError('Negative indices are not supported')
                xargs.append(key.stop)
                if key.step:
                    limit = 1
                    xargs.append(key.step)
                else:
                    limit = self.limit
                    xargs.append(limit)
            else:
                iterator_class = count
                limit = self.limit
                xargs.append(limit)

            s = 0           # Current number of results.
            results = []    # Each result is a dict from JSON data.
            c = None        # 'count' value populated from first response.
            for offset in iterator_class(*xargs):
                request = self.request.clone()    # Copies params and headers.
                request.set_param('offset', offset)
                request.set_param('limit', limit)
                response = request.get()
                results += response['results']
                c = response['count']
                if len(results) - s == 0 or self._stop(offset, limit, c, key.step):
                    break

                s = len(results)
            return [self._result(result) for result in results]

        elif isinstance(key, int):
            if key in self.cache:
                return self.cache[key]

            if int < 0:
                raise IndexError('Negative indices are not supported')
            request = self.request.clone()
            request.set_param('offset', key)
            request.set_param('limit', 1)
            results = request.get()['results']
            if len(results) == 0:
                raise IndexError('Index %i beyond end of results' % key)
            self.cache[key] = self._result(results[0])
            return self.cache[key]
        raise TypeError('')




import os
import cPickle as pickle
from collections import Counter, defaultdict
from uuid import uuid4




identifier_fields = ['doi', 'ayjid', 'issn', 'isbn', 'uri']
paper_fields = [
    ('date', 'publication_date'),
    ('title', 'title'),
    ('volume', 'volume'),
    ('issue', 'issue'),
    ('abstract', 'abstract'),
    ('journal', 'journal'),
]

exclude_fields = [
    'citations',
    'citedReferences'
]


class CorpusHandler(object):
    _add_order = [
        'paper_instance',
        'instance_identifier',
        'instance_metadatum',
        'author_instance',
        'institution_instance',
        'affiliation_instance',
    ]

    def __init__(self, client, tethne_corpus, label, source, batch_size=100, corpus=None):
        self.client = client
        self.id_map = {}
        self.batch_size = batch_size
        self.hoppers = defaultdict(list)

        if not corpus:
            corpus = self.client.create_corpus({
                'source': source,
                'label': label,
            })
        self.corpus = corpus
        self.tethne_corpus = tethne_corpus

    def run(self):
        for tethne_paper in self.tethne_corpus:
            paper_id = self._handle_paper(tethne_paper)

            for tethne_reference in getattr(tethne_paper, 'citedReferences', []):
                if not tethne_reference:
                    continue
                self._handle_cited_reference(tethne_reference, paper_id)

            if len(self.hoppers['paper_instance']) >= self.batch_size:
                self._commit()
        self._commit()

    def _commit(self):
        for model_name in self._add_order:
            instances = []
            for datum in self.hoppers[model_name]:
                for field in ['paper_id', 'author_id', 'institution_id', 'cited_by_id']:
                    if field not in datum:
                        continue
                    ident = datum[field]
                    datum[field] = self.id_map.get(ident, ident)
                instances.append(datum)
            result = self.client.create_bulk(model_name, instances)
            self.id_map.update(result['id_map'])
            print 'Created %i instances of %s' % (len(self.hoppers[model_name]), model_name)
            self.hoppers[model_name] = []

    def _add_instance(self, model_name, instance):
        ident = unicode(uuid4())
        instance.update({'id': ident})
        self.hoppers[model_name].append(instance)
        return ident

    def _exclude_paper_field(self, field):
        exclude = list(zip(*paper_fields)[0]) + identifier_fields + exclude_fields
        return field.startswith('_') or field in exclude

    def _generate_metadata(self, tethne_paper):
        # Generate data for Metadatum model.
        metadata = []

        for field, value in tethne_paper.__dict__.iteritems():
            if self._exclude_paper_field(field):
                continue
            # value = pickle.dumps(value)
            # if type(value) is not unicode:
            #     value = pickle.dumps(value)

            metadata.append({
                'name': field,
                'value': value,
                'corpus_id': self.corpus.id,

            })
        return metadata


    def _generate_identifiers(self, tethne_paper):
        # Generate data for Identifier model.
        identifiers = []
        for field in identifier_fields:
            value = getattr(tethne_paper, field, None)
            if value:
                identifiers.append({
                    'name': field,
                    'value': value,
                    'corpus_id': self.corpus.id,

                })
        return identifiers


    def _handle_paper(self, tethne_paper, **additional):
        paper_data = {}
        for pfield, dbfield in paper_fields:
            value = getattr(tethne_paper, pfield, None)
            if value:
                paper_data[dbfield] = value

        paper_data.update({
            'corpus_id': self.corpus.id,
        })
        paper_data.update(**additional)
        paper_id = self._add_instance('paper_instance', paper_data)

        metadata = []
        for metadata_data in self._generate_metadata(tethne_paper):
            metadata_data.update({'paper_id': paper_id})
            self._add_instance('instance_metadatum', metadata_data)

        identifiers = []
        for identifier_data in self._generate_identifiers(tethne_paper):

            identifier_data.update({'paper_id': paper_id})
            self._add_instance('instance_identifier', identifier_data)

        self._handle_authors(tethne_paper, paper_id)
        return paper_id


    def _handle_cited_reference(self, tethne_reference, paper_id):
        return self._handle_paper(tethne_reference, cited_by_id=paper_id, concrete=False)


    def _handle_institution(self, address, paper_id):
        return self._add_instance('institution_instance', {
            'name': address[0],
            'country': address[1],
            'address': address[2],
            'paper_id': paper_id,
            'corpus_id': self.corpus.id,

        })

    def _handle_authors(self, tethne_paper, paper_id):
        # So that we don't create multiple InstitutionInstances for what are clearly
        #  the same institutions, we first extract all unique address tuples and
        #  create InstitutionInstances before proceeding.
        addresses = getattr(tethne_paper, 'addresses', {})
        normalize = lambda a: (a[0], a[1], ', '.join(a[2]))
        institution_instances = {i: self._handle_institution(i, paper_id) for i
                                 in set([normalize(j) for jset in addresses.values() for j in jset])}

        institutions = {
            name: [institution_instances[normalize(address)] for address in author_addresses]
            for name, author_addresses in addresses.items()
        }

        authors = []
        affiliations = []
        for tethne_author in tethne_paper.authors:
            author_id = self._add_instance('author_instance', {
                'paper_id': paper_id,
                'last_name': tethne_author[0][0],
                'first_name': tethne_author[0][1],
                'corpus_id': self.corpus.id,

            })

            tethne_affiliations = institutions.get(tethne_author,
                                                   institutions.get('__all__',
                                                                    None))
            if not tethne_affiliations:
                continue

            for institution_id in tethne_affiliations:
                self._add_instance('affiliation_instance', {
                    'paper_id': paper_id,
                    'author_id': author_id,
                    'institution_id': institution_id,
                    'confidence': 1./len(tethne_affiliations),
                    'corpus_id': self.corpus.id,
                })



__all__ = ['TethneClient',]

import json, requests, copy, jsonpickle, time
from urlparse import urlparse, parse_qs, urljoin, urlunparse, SplitResult
from urllib import urlencode

from tethneweb.classes import Corpus, Paper
from tethneweb.upload import CorpusHandler
from tethneweb.results import Result, ResultList


class Request(object):
    def __init__(self, client, url, params={}, headers={},
                 get_method=requests.get, post_method=requests.post,
                 retries=3):
        self.url, self.params = self._prep_path(url, params)
        self.headers = headers
        self.client = client
        self.get_method = get_method
        self.post_method = post_method
        self.retries_start = retries
        self.retries = retries

    def set_header(self, key, value):
        self.headers.update({key: value})

    def set_param(self, key, value):
        self.params.update({key: value})

    def _handle_response(self, response):
        if response.status_code != requests.codes.ok:
            raise IOError(response.status_code)
        return response.json()

    def _prep_path(self, path, params={}):
        """
        Pull query parameters from ``path`` and update ``params``.

        Returns
        -------
        tuple
            (url, params)

        """
        o = urlparse(path)
        params.update({k: v[0] for k, v in parse_qs(o.query).iteritems()})
        new = SplitResult(o.scheme, o.netloc, o.path, '', '')
        return new.geturl(), params

    def clone(self):
        return Request(self.client, self.url, params=copy.copy(self.params),
                       headers=copy.copy(self.headers),
                       get_method=self.get_method,
                       post_method=self.post_method,
                       retries=self.retries_start)

    def get(self):
        while self.retries > 0:
            response = self.get_method(self.url, params=self.params, headers=self.headers)
            try:
                return self._handle_response(response)
            except IOError:
                self.retries -= 1
                time.sleep(1)


    def post(self):
        while self.retries > 0:
            response = self.post_method(self.url, data=self.params, headers=self.headers)
            try:
                return self._handle_response(response)
            except IOError:
                self.retries -= 1
                time.sleep(1)



class TethneClient(object):
    def __init__(self, endpoint, username, password, authenticate=True,
                 get_method=requests.get, post_method=requests.post):
        self.endpoint = endpoint
        self.username = username
        self.password = password
        self.get_method = get_method
        self.post_method = post_method
        if authenticate:
            self.authenticate()

    def _methods(self):
        return {'get_method': self.get_method, 'post_method': self.post_method}

    def _path(self, partial):
        return u'/'.join([self.endpoint, partial])
        # return urljoin(self.endpoint, partial)

    def _auth_header(self):
        return {'Authorization': 'Token %s' % self.token}

    def _new_request(self, *args, **kwargs):
        kwargs.update(self._methods())
        return Request(self, *args, **kwargs)

    def _get_or_fail(self, path, params={}, message=None):
        path = path if path.startswith('http') else self._path(path)
        return self._new_request(path, params, self._auth_header())

    def _post_or_fail(self, path, data, with_headers=False):
        path = path if path.startswith('http') else self._path(path)
        headers = self._auth_header() if with_headers else {}
        return self._new_request(path, data, headers)

    def authenticate(self):
        """
        Attempt to retrieve an authentication token.
        """
        auth_data = {'username': self.username, 'password': self.password}
        message = 'Could not authenticate. Please check endpoint and ' \
                + ' credentials, and try again.'

        data = self._post_or_fail('api-token-auth/', auth_data).post()
        self.token = data.get('token', None)

    def follow_link(self, path, result_class=dict, paginated=True, limit=None, **params):
        request = self._new_request(path, params, self._auth_header())
        if paginated:
            return ResultList(request, result_class)
        return Result(request, result_class)

    def list_corpora(self, **params):
        """
        List all corpora to which the user has access.
        """
        return ResultList(self._get_or_fail('rest/corpus/', params=params), Corpus)

    def list_papers(self, **params):
        """
        List all papers to which the user has access.
        """
        return ResultList(self._get_or_fail('rest/paper_instance/', params=params), Paper)

    def list_authors(self, **params):
        """
        List all papers to which the user has access.
        """
        return ResultList(self._get_or_fail('rest/author_instance/', params=params), Author)

    def list_institutions(self, **params):
        """
        List all papers to which the user has access.
        """
        return ResultList(self._get_or_fail('rest/institution_instance/', params=params), Institution)

    def get_paper(self, id):
        return Result(self._get_or_fail('rest/paper_instance/%i/' % int(id)), Paper)

    def get_author(self, id):
        return Result(self._get_or_fail('rest/author_instance/%i/' % int(id)), Author)

    def get_corpus(self, id):
        return Result(self._get_or_fail('rest/corpus/%i/' % int(id)), Corpus)

    def get_institution(self, id):
        return Result(self._get_or_fail('rest/institution_instance/%i/' % int(id)), Institution)

    def upload(self, tethne_corpus, label, source, batch_size=100, corpus=None):
        handler = CorpusHandler(self, tethne_corpus, label, source, batch_size, corpus=corpus)
        handler.run()

    def create_corpus(self, data):
        return Corpus(self, self._post_or_fail('rest/corpus/', data, with_headers=True).post())

    def create_bulk(self, model_name, data):
        return self._post_or_fail('rest/%s/' % model_name, {'data': jsonpickle.encode(data)}, with_headers=True).post()
